# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import re
import textwrap

from twisted.internet import defer
from twisted.internet import reactor
from zope.interface import implementer

from buildbot import util
from buildbot.interfaces import IRenderable
from buildbot.process import buildstep
from buildbot.process import remotecommand
from buildbot.process import results
from buildbot.steps.source.base import Source


@implementer(IRenderable)
class RepoDownloadsFromProperties(util.ComparableMixin):
    parse_download_re = (re.compile(r"repo download ([^ ]+) ([0-9]+/[0-9]+)"),
                         re.compile(r"([^ ]+) ([0-9]+/[0-9]+)"),
                         re.compile(r"([^ ]+)/([0-9]+/[0-9]+)"),
                         )

    compare_attrs = ('names',)

    def __init__(self, names):
        self.names = names

    def getRenderingFor(self, props):
        downloads = []
        for propName in self.names:
            s = props.getProperty(propName)
            if s is not None:
                downloads.extend(self.parseDownloadProperty(s))
        return downloads

    def parseDownloadProperty(self, s):
        """
         lets try to be nice in the format we want
         can support several instances of "repo download proj number/patch"
         (direct copy paste from gerrit web site) or several instances of "proj number/patch"
         (simpler version)
         This feature allows integrator to build with several pending interdependent changes.
         returns list of repo downloads sent to the worker
         """
        if s is None:
            return []
        ret = []
        for cur_re in self.parse_download_re:
            res = cur_re.search(s)
            while res:
                ret.append(f"{res.group(1)} {res.group(2)}")
                s = s[:res.start(0)] + s[res.end(0):]
                res = cur_re.search(s)
        return ret


@implementer(IRenderable)
class RepoDownloadsFromChangeSource(util.ComparableMixin):
    compare_attrs = ('codebase',)

    def __init__(self, codebase=None):
        self.codebase = codebase

    def getRenderingFor(self, props):
        downloads = []
        if self.codebase is None:
            changes = props.getBuild().allChanges()
        else:
            changes = props.getBuild().getSourceStamp(self.codebase).changes
        for change in changes:
            if ("event.type" in change.properties and
                    change.properties["event.type"] == "patchset-created"):
                downloads.append(f'{change.properties["event.change.project"]} '
                                 f'{change.properties["event.change.number"]}/'
                                 f'{change.properties["event.patchSet.number"]}')
        return downloads


class Repo(Source):

    """ Class for Repo with all the smarts """
    name = 'repo'
    renderables = ["manifestURL", "manifestBranch", "manifestFile", "tarball", "jobs",
                   "syncAllBranches", "updateTarballAge", "manifestOverrideUrl",
                   "repoDownloads", "depth", "submodules"]

    ref_not_found_re = re.compile(r"fatal: Couldn't find remote ref")
    cherry_pick_error_re = re.compile(r"|".join([r"Automatic cherry-pick failed",
                                                 r"error: "
                                                 r"fatal: "
                                                 r"possibly due to conflict resolution."]))
    re_change = re.compile(r".* refs/changes/\d\d/(\d+)/(\d+) -> FETCH_HEAD$")
    re_head = re.compile(r"^HEAD is now at ([0-9a-f]+)...")
    # number of retries, if we detect mirror desynchronization
    mirror_sync_retry = 10
    # wait 1min between retries (thus default total retry time is 10min)
    mirror_sync_sleep = 60

    def __init__(self,
                 manifestURL=None,
                 manifestBranch="master",
                 manifestFile="default.xml",
                 tarball=None,
                 jobs=None,
                 syncAllBranches=False,
                 updateTarballAge=7 * 24.0 * 3600.0,
                 manifestOverrideUrl=None,
                 repoDownloads=None,
                 depth=0,
                 submodules=False,
                 syncQuietly=False,
                 **kwargs):
        """
        @type  manifestURL: string
        @param manifestURL: The URL which points at the repo manifests repository.

        @type  manifestBranch: string
        @param manifestBranch: The manifest branch to check out by default.

        @type  manifestFile: string
        @param manifestFile: The manifest to use for sync.

        @type syncAllBranches: bool.
        @param syncAllBranches: true, then we must slowly synchronize all branches.

        @type updateTarballAge: float
        @param updateTarballAge: renderable to determine the update tarball policy,
                                 given properties
                               Returns: max age of tarball in seconds, or None, if we
                               want to skip tarball update

        @type manifestOverrideUrl: string
        @param manifestOverrideUrl: optional http URL for overriding the manifest
                                    usually coming from Property setup by a ForceScheduler

        @type repoDownloads: list of strings
        @param repoDownloads: optional repo download to perform after the repo sync

        @type depth: integer
        @param depth: optional depth parameter to repo init.
                          If specified, create a shallow clone with given depth.

        @type submodules: string
        @param submodules: optional submodules parameter to repo init.

        @type syncQuietly: bool.
        @param syncQuietly: true, then suppress verbose output from repo sync.
        """
        self.manifestURL = manifestURL
        self.manifestBranch = manifestBranch
        self.manifestFile = manifestFile
        self.tarball = tarball
        self.jobs = jobs
        self.syncAllBranches = syncAllBranches
        self.updateTarballAge = updateTarballAge
        self.manifestOverrideUrl = manifestOverrideUrl
        if repoDownloads is None:
            repoDownloads = []
        self.repoDownloads = repoDownloads
        self.depth = depth
        self.submodules = submodules
        self.syncQuietly = syncQuietly
        super().__init__(**kwargs)

        assert self.manifestURL is not None

    def computeSourceRevision(self, changes):
        if not changes:
            return None
        return changes[-1].revision

    def filterManifestPatches(self):
        """
        Patches to manifest projects are a bit special.
        repo does not support a way to download them automatically,
        so we need to implement the boilerplate manually.
        This code separates the manifest patches from the other patches,
        and generates commands to import those manifest patches.
        """
        manifest_unrelated_downloads = []
        manifest_related_downloads = []
        for download in self.repoDownloads:
            project, ch_ps = download.split(" ")[-2:]
            if (self.manifestURL.endswith("/" + project) or
                    self.manifestURL.endswith("/" + project + ".git")):
                ch, ps = map(int, ch_ps.split("/"))
                branch = f"refs/changes/{ch % 100:02}/{ch}/{ps}"
                manifest_related_downloads.append(
                    ["git", "fetch", self.manifestURL, branch])
                manifest_related_downloads.append(
                    ["git", "cherry-pick", "FETCH_HEAD"])
            else:
                manifest_unrelated_downloads.append(download)
        self.repoDownloads = manifest_unrelated_downloads
        self.manifestDownloads = manifest_related_downloads

    def _repoCmd(self, command, abandonOnFailure=True, **kwargs):
        return self._Cmd(["repo"] + command, abandonOnFailure=abandonOnFailure, **kwargs)

    @defer.inlineCallbacks
    def _Cmd(self, command, abandonOnFailure=True, workdir=None, **kwargs):
        if workdir is None:
            workdir = self.workdir
        cmd = remotecommand.RemoteShellCommand(workdir, command,
                                               env=self.env,
                                               logEnviron=self.logEnviron,
                                               timeout=self.timeout, **kwargs)
        self.lastCommand = cmd
        # does not make sense to logEnviron for each command (just for first)
        self.logEnviron = False
        cmd.useLog(self.stdio_log, False)
        yield self.stdio_log.addHeader(f'Starting command: {" ".join(command)}\n')
        self.description = ' '.join(command[:2])
        # FIXME: enable when new style step is switched on yield self.updateSummary()
        yield self.runCommand(cmd)

        if abandonOnFailure and cmd.didFail():
            self.descriptionDone = f'repo failed at: {" ".join(command[:2])}'
            msg = f"Source step failed while running command {cmd}\n"
            yield self.stdio_log.addStderr(msg)
            raise buildstep.BuildStepFailed()
        return cmd.rc

    def repoDir(self):
        return self.build.path_module.join(self.workdir, ".repo")

    def sourcedirIsUpdateable(self):
        return self.pathExists(self.repoDir())

    def run_vc(self, branch, revision, patch):
        return self.doStartVC()

    @defer.inlineCallbacks
    def doStartVC(self):
        self.stdio_log = yield self.addLogForRemoteCommands("stdio")

        self.filterManifestPatches()

        if self.repoDownloads:
            yield self.stdio_log.addHeader("will download:\nrepo download {}\n".format(
                    "\nrepo download ".join(self.repoDownloads)))

        self.willRetryInCaseOfFailure = True

        try:
            yield self.doRepoSync()
        except buildstep.BuildStepFailed as e:
            if not self.willRetryInCaseOfFailure:
                raise
            yield self.stdio_log.addStderr("got issue at first try:\n" + str(e) +
                                           "\nRetry after clobber...")
            yield self.doRepoSync(forceClobber=True)

        yield self.maybeUpdateTarball()

        # starting from here, clobbering will not help
        yield self.doRepoDownloads()
        return results.SUCCESS

    @defer.inlineCallbacks
    def doClobberStart(self):
        yield self.runRmdir(self.workdir)
        yield self.runMkdir(self.workdir)
        yield self.maybeExtractTarball()

    @defer.inlineCallbacks
    def doRepoSync(self, forceClobber=False):
        updatable = yield self.sourcedirIsUpdateable()
        if not updatable or forceClobber:
            # no need to re-clobber in case of failure
            self.willRetryInCaseOfFailure = False
            yield self.doClobberStart()
        yield self.doCleanup()
        command = ['init',
                   '-u', self.manifestURL,
                   '-b', self.manifestBranch,
                   '-m', self.manifestFile,
                   '--depth', str(self.depth)]

        if self.submodules:
            command.append('--submodules')

        yield self._repoCmd(command)

        if self.manifestOverrideUrl:
            msg = f"overriding manifest with {self.manifestOverrideUrl}\n"
            yield self.stdio_log.addHeader(msg)

            local_path = self.build.path_module.join(self.workdir, self.manifestOverrideUrl)
            local_file = yield self.pathExists(local_path)
            if local_file:
                yield self._Cmd(["cp", "-f", self.manifestOverrideUrl, "manifest_override.xml"])
            else:
                yield self._Cmd(["wget", self.manifestOverrideUrl, "-O", "manifest_override.xml"])
            yield self._Cmd(["ln", "-sf", "../manifest_override.xml", "manifest.xml"],
                            workdir=self.build.path_module.join(self.workdir, ".repo"))

        for command in self.manifestDownloads:
            yield self._Cmd(command, workdir=self.build.path_module.join(self.workdir, ".repo",
                                                                         "manifests"))

        command = ['sync', '--force-sync']
        if self.jobs:
            command.append('-j' + str(self.jobs))
        if not self.syncAllBranches:
            command.append('-c')
        if self.syncQuietly:
            command.append('-q')
        self.description = "repo sync"
        # FIXME: enable when new style step is used: yield self.updateSummary()
        yield self.stdio_log.addHeader(f"synching manifest {self.manifestFile} from branch "
                                       f"{self.manifestBranch} from {self.manifestURL}\n")
        yield self._repoCmd(command)

        command = ['manifest', '-r', '-o', 'manifest-original.xml']
        yield self._repoCmd(command)

    # check whether msg matches one of the
    # compiled regexps in self.re_error_messages
    def _findErrorMessages(self, error_re):
        for logname in ['stderr', 'stdout']:
            if not hasattr(self.lastCommand, logname):
                continue
            msg = getattr(self.lastCommand, logname)
            if not re.search(error_re, msg) is None:
                return True
        return False

    def _sleep(self, delay):
        d = defer.Deferred()
        reactor.callLater(delay, d.callback, 1)
        return d

    @defer.inlineCallbacks
    def doRepoDownloads(self):
        self.repo_downloaded = ""
        for download in self.repoDownloads:
            command = ['download'] + download.split(' ')
            yield self.stdio_log.addHeader(f"downloading changeset {download}\n")

            retry = self.mirror_sync_retry + 1
            while retry > 0:
                yield self._repoCmd(command, abandonOnFailure=False,
                                    collectStdout=True, collectStderr=True)
                if not self._findErrorMessages(self.ref_not_found_re):
                    break
                retry -= 1
                yield self.stdio_log.addStderr(f"failed downloading changeset {download}\n")
                yield self.stdio_log.addHeader("wait one minute for mirror sync\n")
                yield self._sleep(self.mirror_sync_sleep)

            if retry == 0:
                self.descriptionDone = f"repo: change {download} does not exist"
                raise buildstep.BuildStepFailed()

            if self.lastCommand.didFail() or self._findErrorMessages(self.cherry_pick_error_re):
                # cherry pick error! We create a diff with status current workdir
                # in stdout, which reveals the merge errors and exit
                command = ['forall', '-c', 'git', 'diff', 'HEAD']
                yield self._repoCmd(command, abandonOnFailure=False)
                self.descriptionDone = f"download failed: {download}"
                raise buildstep.BuildStepFailed()

            if hasattr(self.lastCommand, 'stderr'):
                lines = self.lastCommand.stderr.split("\n")
                match1 = match2 = False
                for line in lines:
                    if not match1:
                        match1 = self.re_change.match(line)
                    if not match2:
                        match2 = self.re_head.match(line)
                if match1 and match2:
                    self.repo_downloaded += (f"{match1.group(1)}/{match1.group(2)} "
                                             f"{match2.group(1)} ")

        self.setProperty("repo_downloaded", self.repo_downloaded, "Source")

    def computeTarballOptions(self):
        # Keep in mind that the compression part of tarball generation
        # can be non negligible
        tar = ['tar']
        if self.tarball.endswith("pigz"):
            tar.append('-I')
            tar.append('pigz')
        elif self.tarball.endswith("gz"):
            tar.append('-z')
        elif self.tarball.endswith("bz2") or self.tarball.endswith("bz"):
            tar.append('-j')
        elif self.tarball.endswith("lzma"):
            tar.append('--lzma')
        elif self.tarball.endswith("lzop"):
            tar.append('--lzop')
        return tar

    @defer.inlineCallbacks
    def maybeExtractTarball(self):
        if self.tarball:
            tar = self.computeTarballOptions() + ['-xvf', self.tarball]
            res = yield self._Cmd(tar, abandonOnFailure=False)
            if res:  # error with tarball.. erase repo dir and tarball
                yield self._Cmd(["rm", "-f", self.tarball], abandonOnFailure=False)
                yield self.runRmdir(self.repoDir(), abandonOnFailure=False)

    @defer.inlineCallbacks
    def maybeUpdateTarball(self):
        if not self.tarball or self.updateTarballAge is None:
            return
        # tarball path is absolute, so we cannot use worker's stat command
        # stat -c%Y gives mtime in second since epoch
        res = yield self._Cmd(["stat", "-c%Y", self.tarball], collectStdout=True,
                              abandonOnFailure=False)
        if not res:
            tarball_mtime = int(self.lastCommand.stdout)
            yield self._Cmd(["stat", "-c%Y", "."], collectStdout=True)
            now_mtime = int(self.lastCommand.stdout)
            age = now_mtime - tarball_mtime
        if res or age > self.updateTarballAge:
            tar = self.computeTarballOptions() + \
                ['-cvf', self.tarball, ".repo"]
            res = yield self._Cmd(tar, abandonOnFailure=False)
            if res:  # error with tarball.. erase tarball, but don't fail
                yield self._Cmd(["rm", "-f", self.tarball], abandonOnFailure=False)

    # a simple shell script to gather all cleanup tweaks...
    # doing them one by one just complicate the stuff
    # and mess up the stdio log
    def _getCleanupCommand(self):
        """also used by tests for expectations"""
        return textwrap.dedent("""\
            set -v
            if [ -d .repo/manifests ]
            then
                # repo just refuse to run if manifest is messed up
                # so ensure we are in a known state
                cd .repo/manifests
                rm -f .git/index.lock
                git fetch origin
                git reset --hard remotes/origin/%(manifestBranch)s
                git config branch.default.merge %(manifestBranch)s
                cd ..
                ln -sf manifests/%(manifestFile)s manifest.xml
                cd ..
             fi
             repo forall -c rm -f .git/index.lock
             repo forall -c git clean -f -d -x 2>/dev/null
             repo forall -c git reset --hard HEAD 2>/dev/null
             rm -f %(workdir)s/.repo/project.list
             """) % dict(manifestBranch=self.manifestBranch,
                         manifestFile=self.manifestFile,
                         workdir=self.workdir)

    def doCleanup(self):
        command = self._getCleanupCommand()
        return self._Cmd(["bash", "-c", command], abandonOnFailure=False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import build_data
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces


class TestBuildDataNoValueEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = build_data.BuildDataNoValueEndpoint
    resourceTypeClass = build_data.BuildData

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88,
                         workerid=47),
            fakedb.BuildData(id=91, buildid=30, name='name1', value=b'value1', source='source1'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_build_id(self):
        result = yield self.callGet(('builds', 30, 'data', 'name1'))
        self.validateData(result)
        self.assertEqual(result, {
            'buildid': 30,
            'name': 'name1',
            'value': None,
            'source': 'source1',
            'length': 6,
        })

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_builder_name_build_number(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 7, 'data', 'name1'))
        self.validateData(result)
        self.assertEqual(result, {
            'buildid': 30,
            'name': 'name1',
            'value': None,
            'source': 'source1',
            'length': 6,
        })

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_builder_id_build_number(self):
        result = yield self.callGet(('builders', 88, 'builds', 7, 'data', 'name1'))
        self.validateData(result)
        self.assertEqual(result, {
            'buildid': 30,
            'name': 'name1',
            'value': None,
            'length': 6,
            'source': 'source1',
        })

    @defer.inlineCallbacks
    def test_get_missing_by_build_id_missing_build(self):
        result = yield self.callGet(('builds', 31, 'data', 'name1'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_build_id_missing_name(self):
        result = yield self.callGet(('builds', 30, 'data', 'name_missing'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_builder(self):
        result = yield self.callGet(('builders', 'b_missing', 'builds', 7, 'data', 'name1'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_build(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 17, 'data', 'name1'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_name(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 7, 'data', 'name_missing'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_builder(self):
        result = yield self.callGet(('builders', 188, 'builds', 7, 'data', 'name1'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_build(self):
        result = yield self.callGet(('builders', 88, 'builds', 17, 'data', 'name1'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_name(self):
        result = yield self.callGet(('builders', 88, 'builds', 7, 'data', 'name_missing'))
        self.assertIsNone(result)


class TestBuildDataEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = build_data.BuildDataEndpoint
    resourceTypeClass = build_data.BuildData

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88,
                         workerid=47),
            fakedb.BuildData(id=91, buildid=30, name='name1', value=b'value1', source='source1'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    def validateData(self, data):
        self.assertIsInstance(data['raw'], bytes)
        self.assertIsInstance(data['mime-type'], str)
        self.assertIsInstance(data['filename'], str)

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_build_id(self):
        result = yield self.callGet(('builds', 30, 'data', 'name1', 'value'))
        self.validateData(result)
        self.assertEqual(result, {
            'raw': b'value1',
            'mime-type': 'application/octet-stream',
            'filename': 'name1',
        })

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_builder_name_build_number(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 7, 'data', 'name1', 'value'))
        self.validateData(result)
        self.assertEqual(result, {
            'raw': b'value1',
            'mime-type': 'application/octet-stream',
            'filename': 'name1',
        })

    @defer.inlineCallbacks
    def test_get_existing_build_data_by_builder_id_build_number(self):
        result = yield self.callGet(('builders', 88, 'builds', 7, 'data', 'name1', 'value'))
        self.validateData(result)
        self.assertEqual(result, {
            'raw': b'value1',
            'mime-type': 'application/octet-stream',
            'filename': 'name1',
        })

    @defer.inlineCallbacks
    def test_get_missing_by_build_id_missing_build(self):
        result = yield self.callGet(('builds', 31, 'data', 'name1', 'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_build_id_missing_name(self):
        result = yield self.callGet(('builds', 30, 'data', 'name_missing', 'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_builder(self):
        result = yield self.callGet(('builders', 'b_missing', 'builds', 7, 'data', 'name1',
                                     'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_build(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 17, 'data', 'name1', 'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_name_build_number_missing_name(self):
        result = yield self.callGet(('builders', 'b1', 'builds', 7, 'data', 'name_missing',
                                     'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_builder(self):
        result = yield self.callGet(('builders', 188, 'builds', 7, 'data', 'name1', 'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_build(self):
        result = yield self.callGet(('builders', 88, 'builds', 17, 'data', 'name1', 'value'))
        self.assertIsNone(result)

    @defer.inlineCallbacks
    def test_get_missing_by_builder_id_build_number_missing_name(self):
        result = yield self.callGet(('builders', 88, 'builds', 7, 'data', 'name_missing', 'value'))
        self.assertIsNone(result)


class TestBuildDatasNoValueEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = build_data.BuildDatasNoValueEndpoint
    resourceTypeClass = build_data.BuildData

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.Master(id=88),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.BuildRequest(id=42, buildsetid=20, builderid=88),
            fakedb.BuildRequest(id=43, buildsetid=20, builderid=88),
            fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88,
                         workerid=47),
            fakedb.Build(id=31, buildrequestid=42, number=8, masterid=88, builderid=88,
                         workerid=47),
            fakedb.Build(id=32, buildrequestid=42, number=9, masterid=88, builderid=88,
                         workerid=47),
            fakedb.BuildData(id=91, buildid=30, name='name1', value=b'value1', source='source1'),
            fakedb.BuildData(id=92, buildid=30, name='name2', value=b'value2', source='source2'),
            fakedb.BuildData(id=93, buildid=31, name='name3', value=b'value3', source='source3'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @parameterized.expand([
        ('multiple_values', 7, ['name1', 'name2']),
        ('single_value', 8, ['name3']),
        ('no_values', 9, []),
        ('not_existing', 10, []),
    ])
    @defer.inlineCallbacks
    def test_get_builders_builder_name(self, name, build_number, exp_names):
        results = yield self.callGet(('builders', 'b1', 'builds', build_number, 'data'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['name'] for r in results], exp_names)

    @parameterized.expand([
        ('multiple_values', 7, ['name1', 'name2']),
        ('single_value', 8, ['name3']),
        ('no_values', 9, []),
        ('not_existing', 10, []),
    ])
    @defer.inlineCallbacks
    def test_get_builders_builder_id(self, name, build_number, exp_names):
        results = yield self.callGet(('builders', 88, 'builds', build_number, 'data'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['name'] for r in results], exp_names)

    @parameterized.expand([
        ('multiple_values', 30, ['name1', 'name2']),
        ('single_value', 31, ['name3']),
        ('no_values', 32, []),
        ('not_existing', 33, []),
    ])
    @defer.inlineCallbacks
    def test_get_builds_id(self, name, buildid, exp_names):
        results = yield self.callGet(('builds', buildid, 'data'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['name'] for r in results], exp_names)


class TestBuildData(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = build_data.BuildData(self.master)

    def test_signature_set_build_data(self):
        @self.assertArgSpecMatches(self.master.data.updates.setBuildData,
                                   self.rtype.setBuildData)
        def setBuildData(self, buildid, name, value, source):
            pass

    @defer.inlineCallbacks
    def test_set_build_data(self):
        yield self.rtype.setBuildData(buildid=2, name='name1', value=b'value1', source='source1')

        result = yield self.master.db.build_data.getBuildData(2, 'name1')
        self.assertEqual(result, {
            'buildid': 2,
            'name': 'name1',
            'value': b'value1',
            'length': 6,
            'source': 'source1',
        })



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.test.fakedb.base import FakeDBComponent
from buildbot.test.fakedb.row import Row
from buildbot.util import epoch2datetime


class Master(Row):
    table = "masters"

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]

    def __init__(self, id=None, name='some:master', name_hash=None, active=1, last_active=9998999):
        super().__init__(id=id, name=name, name_hash=name_hash, active=active,
                         last_active=last_active)


class FakeMastersComponent(FakeDBComponent):

    data2db = {"masterid": "id", "link": "id"}

    def setUp(self):
        self.masters = {}

    def insertTestData(self, rows):
        for row in rows:
            if isinstance(row, Master):
                self.masters[row.id] = dict(
                    id=row.id,
                    name=row.name,
                    active=bool(row.active),
                    last_active=epoch2datetime(row.last_active))

    def findMasterId(self, name):
        for m in self.masters.values():
            if m['name'] == name:
                return defer.succeed(m['id'])
        id = len(self.masters) + 1
        self.masters[id] = dict(
            id=id,
            name=name,
            active=False,
            last_active=epoch2datetime(self.reactor.seconds()))
        return defer.succeed(id)

    def setMasterState(self, masterid, active):
        if masterid in self.masters:
            was_active = self.masters[masterid]['active']
            self.masters[masterid]['active'] = active
            if active:
                self.masters[masterid]['last_active'] = \
                    epoch2datetime(self.reactor.seconds())
            return defer.succeed(bool(was_active) != bool(active))
        else:
            return defer.succeed(False)

    def getMaster(self, masterid):
        if masterid in self.masters:
            return defer.succeed(self.masters[masterid])
        return defer.succeed(None)

    def getMasters(self):
        return defer.succeed(sorted(self.masters.values(),
                                    key=lambda x: x['id']))

    # test helpers

    def markMasterInactive(self, masterid):
        if masterid in self.masters:
            self.masters[masterid]['active'] = False
        return defer.succeed(None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import json

import sqlalchemy as sa

from twisted.internet import defer

from buildbot.db import base
from buildbot.util import epoch2datetime


class StepsConnectorComponent(base.DBConnectorComponent):
    # Documentation is in developer/db.rst
    url_lock = None

    @defer.inlineCallbacks
    def getStep(self, stepid=None, buildid=None, number=None, name=None):
        tbl = self.db.model.steps
        if stepid is not None:
            wc = (tbl.c.id == stepid)
        else:
            if buildid is None:
                raise RuntimeError('must supply either stepid or buildid')
            if number is not None:
                wc = (tbl.c.number == number)
            elif name is not None:
                wc = (tbl.c.name == name)
            else:
                raise RuntimeError('must supply either number or name')
            wc = wc & (tbl.c.buildid == buildid)

        def thd(conn):
            q = self.db.model.steps.select(whereclause=wc)
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._stepdictFromRow(row)
            res.close()
            return rv
        return (yield self.db.pool.do(thd))

    # returns a Deferred that returns a value
    def getSteps(self, buildid):
        def thd(conn):
            tbl = self.db.model.steps
            q = tbl.select()
            q = q.where(tbl.c.buildid == buildid)
            q = q.order_by(tbl.c.number)
            res = conn.execute(q)
            return [self._stepdictFromRow(row) for row in res.fetchall()]
        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def addStep(self, buildid, name, state_string):
        def thd(conn):
            tbl = self.db.model.steps
            # get the highest current number
            r = conn.execute(sa.select([sa.func.max(tbl.c.number)],
                                       whereclause=(tbl.c.buildid == buildid)))
            number = r.scalar()
            number = 0 if number is None else number + 1

            # note that there is no chance for a race condition here,
            # since only one master is inserting steps.  If there is a
            # conflict, then the name is likely already taken.
            insert_row = dict(buildid=buildid, number=number,
                              started_at=None, complete_at=None,
                              state_string=state_string,
                              urls_json='[]', name=name)
            try:
                r = conn.execute(self.db.model.steps.insert(), insert_row)
                got_id = r.inserted_primary_key[0]
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                got_id = None

            if got_id:
                return (got_id, number, name)

            # we didn't get an id, so calculate a unique name and use that
            # instead.  Because names are truncated at the right to fit in a
            # 50-character identifier, this isn't a simple query.
            res = conn.execute(sa.select([tbl.c.name],
                                         whereclause=((tbl.c.buildid == buildid))))
            names = {row[0] for row in res}
            num = 1
            while True:
                numstr = f'_{num}'
                newname = name[:50 - len(numstr)] + numstr
                if newname not in names:
                    break
                num += 1
            insert_row['name'] = newname
            r = conn.execute(self.db.model.steps.insert(), insert_row)
            got_id = r.inserted_primary_key[0]
            return (got_id, number, newname)
        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def startStep(self, stepid):
        started_at = int(self.master.reactor.seconds())

        def thd(conn):
            tbl = self.db.model.steps
            q = tbl.update(whereclause=(tbl.c.id == stepid))
            conn.execute(q, started_at=started_at)
        yield self.db.pool.do(thd)

    # returns a Deferred that returns None
    def setStepStateString(self, stepid, state_string):
        def thd(conn):
            tbl = self.db.model.steps
            q = tbl.update(whereclause=(tbl.c.id == stepid))
            conn.execute(q, state_string=state_string)
        return self.db.pool.do(thd)

    def addURL(self, stepid, name, url, _racehook=None):
        # This methods adds an URL to the db
        # This is a read modify write and thus there is a possibility
        # that several urls are added at the same time (e.g with a deferredlist
        # at the end of a step)
        # this race condition is only inside the same master, as only one master
        # is supposed to add urls to a buildstep.
        # so threading.lock is used, as we are in the thread pool
        if self.url_lock is None:
            # this runs in reactor thread, so no race here..
            self.url_lock = defer.DeferredLock()

        def thd(conn):

            tbl = self.db.model.steps
            wc = (tbl.c.id == stepid)
            q = sa.select([tbl.c.urls_json],
                          whereclause=wc)
            res = conn.execute(q)
            row = res.fetchone()
            if _racehook is not None:
                _racehook()
            urls = json.loads(row.urls_json)

            url_item = dict(name=name, url=url)

            if url_item not in urls:
                urls.append(url_item)
                q = tbl.update(whereclause=wc)
                conn.execute(q, urls_json=json.dumps(urls))

        return self.url_lock.run(lambda: self.db.pool.do(thd))

    # returns a Deferred that returns None
    def finishStep(self, stepid, results, hidden):
        def thd(conn):
            tbl = self.db.model.steps
            q = tbl.update(whereclause=(tbl.c.id == stepid))
            conn.execute(q,
                         complete_at=int(self.master.reactor.seconds()),
                         results=results,
                         hidden=1 if hidden else 0)
        return self.db.pool.do(thd)

    def _stepdictFromRow(self, row):
        return dict(
            id=row.id,
            number=row.number,
            name=row.name,
            buildid=row.buildid,
            started_at=epoch2datetime(row.started_at),
            complete_at=epoch2datetime(row.complete_at),
            state_string=row.state_string,
            results=row.results,
            urls=json.loads(row.urls_json),
            hidden=bool(row.hidden))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime
import textwrap

from dateutil.tz import tzutc

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.reporters import utils
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import logging


class TestDataUtils(TestReactorMixin, unittest.TestCase, logging.LoggingMixin):
    LOGCONTENT = textwrap.dedent("""\
        line zero
        line 1
        """)

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantData=True, wantDb=True,
                                             wantMq=True)

    def setupDb(self):
        self.db = self.master.db
        self.db.insertTestData([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=98, results=SUCCESS, reason="testReason1"),
            fakedb.Buildset(id=99, results=SUCCESS, reason="testReason2", parent_buildid=21),
            fakedb.Builder(id=80, name='Builder1'),
            fakedb.Builder(id=81, name='Builder2'),
            fakedb.BuildRequest(id=9, buildsetid=97, builderid=80),
            fakedb.BuildRequest(id=10, buildsetid=97, builderid=80),
            fakedb.BuildRequest(id=11, buildsetid=98, builderid=80),
            fakedb.BuildRequest(id=12, buildsetid=98, builderid=80),
            fakedb.BuildRequest(id=13, buildsetid=99, builderid=81),
            fakedb.Build(id=18, number=0, builderid=80, buildrequestid=9, workerid=13,
                         masterid=92, results=FAILURE),
            fakedb.Build(id=19, number=1, builderid=80, buildrequestid=10, workerid=13,
                         masterid=92, results=RETRY),
            fakedb.Build(id=20, number=2, builderid=80, buildrequestid=11, workerid=13,
                         masterid=92, results=SUCCESS),
            fakedb.Build(id=21, number=3, builderid=80, buildrequestid=12, workerid=13,
                         masterid=92, results=SUCCESS),
            fakedb.Build(id=22, number=1, builderid=81, buildrequestid=13, workerid=13,
                         masterid=92, results=SUCCESS),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.SourceStamp(id=234),
            fakedb.Change(changeid=13, branch='trunk', revision='9283', author='me@foo',
                          repository='svn://...', codebase='cbsvn',
                          project='world-domination', sourcestampid=234),
            fakedb.Patch(id=99, patch_base64='aGVsbG8sIHdvcmxk',
                         patch_author='him@foo', patch_comment='foo', subdir='/foo',
                         patchlevel=3),
            fakedb.SourceStamp(id=235, patchid=99),
        ])
        for _id in (20, 21):
            self.db.insertTestData([
                fakedb.BuildProperty(
                    buildid=_id, name="workername", value="wrk"),
                fakedb.BuildProperty(
                    buildid=_id, name="reason", value="because"),
                fakedb.BuildProperty(
                    buildid=_id, name="owner", value="him"),
                fakedb.Step(id=100 + _id, buildid=_id, name="step1"),
                fakedb.Step(id=200 + _id, buildid=_id, name="step2"),
                fakedb.Log(id=60 + _id, stepid=100 + _id, name='stdio', slug='stdio', type='s',
                           num_lines=2),
                fakedb.LogChunk(logid=60 + _id, first_line=0, last_line=1, compressed=0,
                                content=self.LOGCONTENT),
            ])

        @defer.inlineCallbacks
        def getChangesForBuild(buildid):
            assert buildid == 20
            ch = yield self.master.db.changes.getChange(13)
            return [ch]

        self.master.db.changes.getChangesForBuild = getChangesForBuild

    @defer.inlineCallbacks
    def test_getDetailsForBuildset(self):
        self.setupDb()
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True,
                                                want_steps=True, want_previous_build=True)
        self.assertEqual(len(res['builds']), 2)
        build1 = res['builds'][0]
        build2 = res['builds'][1]
        buildset = res['buildset']
        self.assertEqual(build1['properties'], {'reason': ('because', 'fakedb'),
                                                'owner': ('him', 'fakedb'),
                                                'workername': ('wrk', 'fakedb')})
        self.assertEqual(len(build1['steps']), 2)
        self.assertEqual(build1['buildid'], 20)
        self.assertEqual(build2['buildid'], 21)
        self.assertEqual(buildset['bsid'], 98)

        # make sure prev_build was computed
        self.assertEqual(build1['prev_build']['buildid'], 18)
        self.assertEqual(build2['prev_build']['buildid'], 20)

    @defer.inlineCallbacks
    def test_getDetailsForBuild(self):
        self.setupDb()
        build = yield self.master.data.get(("builds", 21))
        yield utils.getDetailsForBuild(self.master, build, want_properties=False,
                                       want_steps=False, want_previous_build=False,
                                       want_logs=False)

        self.assertEqual(build['parentbuild'], None)
        self.assertEqual(build['parentbuilder'], None)

    @defer.inlineCallbacks
    def test_getDetailsForBuildWithParent(self):
        self.setupDb()
        build = yield self.master.data.get(("builds", 22))
        yield utils.getDetailsForBuild(self.master, build, want_properties=False,
                                       want_steps=False, want_previous_build=False,
                                       want_logs=False)

        self.assertEqual(build['parentbuild']['buildid'], 21)
        self.assertEqual(build['parentbuilder']['name'], "Builder1")

    @defer.inlineCallbacks
    def test_getDetailsForBuildsetWithLogs(self):
        self.setupDb()
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True,
                                                want_steps=True, want_previous_build=True,
                                                want_logs=True, want_logs_content=True)

        build1 = res['builds'][0]
        self.assertEqual(build1['steps'][0]['logs'][0]['content']['content'], self.LOGCONTENT)
        self.assertEqual(build1['steps'][0]['logs'][0]['url'],
                         'http://localhost:8080/#builders/80/builds/2/steps/29/logs/stdio')

    @defer.inlineCallbacks
    def test_get_details_for_buildset_all(self):
        self.setupDb()
        res = yield utils.getDetailsForBuildset(self.master, 98, want_properties=True,
                                                want_steps=True, want_previous_build=True,
                                                want_logs=True, want_logs_content=True)

        self.assertEqual(res, {
            'builds': [{
                'builder': {
                    'builderid': 80,
                    'description': None,
                    'masterids': [],
                    'name': 'Builder1',
                    'tags': []
                },
                'builderid': 80,
                'buildid': 20,
                'buildrequestid': 11,
                'buildset': {
                    'bsid': 98,
                    'complete': False,
                    'complete_at': None,
                    'external_idstring': 'extid',
                    'parent_buildid': None,
                    'parent_relationship': None,
                    'reason': 'testReason1',
                    'results': 0,
                    'sourcestamps': [{
                        'branch': 'master',
                        'codebase': '',
                        'created_at': datetime.datetime(1972, 11, 5, 18, 7, 14, tzinfo=tzutc()),
                        'patch': None,
                        'project': 'proj',
                        'repository': 'repo',
                        'revision': 'abcd',
                        'ssid': 234
                    }],
                    'submitted_at': 12345678
                },
                'complete': False,
                'complete_at': None,
                'masterid': 92,
                'number': 2,
                'prev_build': {
                    'builderid': 80,
                    'buildid': 18,
                    'buildrequestid': 9,
                    'complete': False,
                    'complete_at': None,
                    'masterid': 92,
                    'number': 0,
                    'properties': {},
                    'results': 2,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': 'test',
                    'workerid': 13
                },
                'properties': {
                    'owner': ('him', 'fakedb'),
                    'reason': ('because', 'fakedb'),
                    'workername': ('wrk', 'fakedb')
                },
                'results': 0,
                'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                'state_string': 'test',
                'steps': [{
                    'buildid': 20,
                    'complete': False,
                    'complete_at': None,
                    'hidden': False,
                    'logs': [{
                        'complete': False,
                        'content': {
                            'content': 'line zero\nline 1\n',
                            'firstline': 0,
                            'logid': 80
                        },
                        'logid': 80,
                        'name': 'stdio',
                        'num_lines': 2,
                        'slug': 'stdio',
                        'stepid': 120,
                        'type': 's',
                        'url': 'http://localhost:8080/#builders/80/builds/2/steps/29/logs/stdio'
                    }],
                    'name': 'step1',
                    'number': 29,
                    'results': None,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': '',
                    'stepid': 120,
                    'urls': []
                }, {
                    'buildid': 20,
                    'complete': False,
                    'complete_at': None,
                    'hidden': False,
                    'logs': [],
                    'name': 'step2',
                    'number': 29,
                    'results': None,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': '',
                    'stepid': 220,
                    'urls': []
                }],
                'url': 'http://localhost:8080/#builders/80/builds/2',
                'workerid': 13
            }, {
                'builder': {
                    'builderid': 80,
                    'description': None,
                    'masterids': [],
                    'name': 'Builder1',
                    'tags': []
                },
                'builderid': 80,
                'buildid': 21,
                'buildrequestid': 12,
                'buildset': {
                    'bsid': 98,
                    'complete': False,
                    'complete_at': None,
                    'external_idstring': 'extid',
                    'parent_buildid': None,
                    'parent_relationship': None,
                    'reason': 'testReason1',
                    'results': 0,
                    'sourcestamps': [{
                        'branch': 'master',
                        'codebase': '',
                        'created_at': datetime.datetime(1972, 11, 5, 18, 7, 14, tzinfo=tzutc()),
                        'patch': None,
                        'project': 'proj',
                        'repository': 'repo',
                        'revision': 'abcd',
                        'ssid': 234
                    }],
                    'submitted_at': 12345678
                },
                'complete': False,
                'complete_at': None,
                'masterid': 92,
                'number': 3,
                'prev_build': {
                    'builderid': 80,
                'buildid': 20,
                    'buildrequestid': 11,
                    'complete': False,
                    'complete_at': None,
                    'masterid': 92,
                    'number': 2,
                    'properties': {},
                    'results': 0,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': 'test',
                    'workerid': 13
                },
                'properties': {
                    'owner': ('him', 'fakedb'),
                    'reason': ('because', 'fakedb'),
                    'workername': ('wrk', 'fakedb')
                },
                'results': 0,
                'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                'state_string': 'test',
                'steps': [{
                    'buildid': 21,
                    'complete': False,
                    'complete_at': None,
                    'hidden': False,
                    'logs': [{
                        'complete': False,
                        'content': {'content': 'line zero\nline 1\n',
                        'firstline': 0,
                        'logid': 81},
                        'logid': 81,
                        'name': 'stdio',
                        'num_lines': 2,
                        'slug': 'stdio',
                        'stepid': 121,
                        'type': 's',
                        'url': 'http://localhost:8080/#builders/80/builds/3/steps/29/logs/stdio'
                    }],
                    'name': 'step1',
                    'number': 29,
                    'results': None,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': '',
                    'stepid': 121,
                    'urls': []
                }, {
                    'buildid': 21,
                    'complete': False,
                    'complete_at': None,
                    'hidden': False,
                    'logs': [],
                    'name': 'step2',
                    'number': 29,
                    'results': None,
                    'started_at': datetime.datetime(2011, 5, 1, 15, 3, 42, tzinfo=tzutc()),
                    'state_string': '',
                    'stepid': 221,
                    'urls': []
                }],
                'url': 'http://localhost:8080/#builders/80/builds/3',
                'workerid': 13
            }],
            'buildset': {
                'bsid': 98,
                'complete': False,
                'complete_at': None,
                'external_idstring': 'extid',
                'parent_buildid': None,
                'parent_relationship': None,
                'reason': 'testReason1',
                'results': 0,
                'sourcestamps': [{
                    'branch': 'master',
                    'codebase': '',
                    'created_at': datetime.datetime(1972, 11, 5, 18, 7, 14, tzinfo=tzutc()),
                    'patch': None,
                    'project': 'proj',
                    'repository': 'repo',
                    'revision': 'abcd',
                    'ssid': 234
                }],
                'submitted_at': 12345678
            }
        })

    @defer.inlineCallbacks
    def test_getResponsibleUsers(self):
        self.setupDb()
        res = yield utils.getResponsibleUsersForSourceStamp(self.master, 234)
        self.assertEqual(res, ["me@foo"])

    @defer.inlineCallbacks
    def test_getResponsibleUsersFromPatch(self):
        self.setupDb()
        res = yield utils.getResponsibleUsersForSourceStamp(self.master, 235)
        self.assertEqual(res, ["him@foo"])

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuild(self):
        self.setupDb()
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertEqual(sorted(res), sorted(["me@foo", "him"]))

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuildWithBadOwner(self):
        self.setUpLogging()
        self.setupDb()
        self.db.insertTestData([
            fakedb.BuildProperty(
                buildid=20, name="owner", value=["him"]),
        ])
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertLogged("Please report a bug")
        self.assertEqual(sorted(res), sorted(["me@foo", "him"]))

    @defer.inlineCallbacks
    def test_getResponsibleUsersForBuildWithOwners(self):
        self.setupDb()
        self.db.insertTestData([
            fakedb.BuildProperty(
                buildid=20, name="owners", value=["him", "her"]),
        ])
        res = yield utils.getResponsibleUsersForBuild(self.master, 20)
        self.assertEqual(sorted(res), sorted(["me@foo", "him", "her"]))

    @defer.inlineCallbacks
    def test_getPreviousBuild(self):
        self.setupDb()
        build = yield self.master.data.get(("builds", 21))
        res = yield utils.getPreviousBuild(self.master, build)
        self.assertEqual(res['buildid'], 20)

    @defer.inlineCallbacks
    def test_getPreviousBuildWithRetry(self):
        self.setupDb()
        build = yield self.master.data.get(("builds", 20))
        res = yield utils.getPreviousBuild(self.master, build)
        self.assertEqual(res['buildid'], 18)


class TestURLUtils(TestReactorMixin, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self)

    def test_UrlForBuild(self):
        self.assertEqual(utils.getURLForBuild(self.master, 1, 3),
                         'http://localhost:8080/#builders/1/builds/3')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import enum

from twisted.internet import defer
from twisted.python.filepath import FilePath
from twisted.trial.unittest import SkipTest

from buildbot.test.fake.worker import SeverWorkerConnectionMixin
from buildbot.test.fake.worker import disconnect_master_side_worker
from buildbot.worker import AbstractLatentWorker

try:
    from buildbot_worker.bot import LocalWorker as RemoteWorker
    from buildbot_worker.pb import BotPbLike
except ImportError:
    RemoteWorker = None


class States(enum.Enum):
    STOPPED = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3


class LatentController(SeverWorkerConnectionMixin):

    """
    A controller for ``ControllableLatentWorker``.

    https://glyph.twistedmatrix.com/2015/05/separate-your-fakes-and-your-inspectors.html

    Note that by default workers will connect automatically if True is passed
    to start_instance().

    Also by default workers will disconnect automatically just as
    stop_instance() is executed.
    """

    def __init__(self, case, name, kind=None, build_wait_timeout=600,
                 starts_without_substantiate=None, **kwargs):
        self.case = case
        self.build_wait_timeout = build_wait_timeout
        self.worker = ControllableLatentWorker(name, self, **kwargs)
        self.remote_worker = None

        if starts_without_substantiate is not None:
            self.worker.starts_without_substantiate = \
                starts_without_substantiate

        self.state = States.STOPPED
        self.auto_stop_flag = False
        self.auto_start_flag = False
        self.auto_connect_worker = True
        self.auto_disconnect_worker = True

        self.kind = kind
        self._started_kind = None
        self._started_kind_deferred = None

    @property
    def starting(self):
        return self.state == States.STARTING

    @property
    def started(self):
        return self.state == States.STARTED

    @property
    def stopping(self):
        return self.state == States.STOPPING

    @property
    def stopped(self):
        return self.state == States.STOPPED

    def auto_start(self, result):
        self.auto_start_flag = result
        if self.auto_start_flag and self.state == States.STARTING:
            self.start_instance(True)

    @defer.inlineCallbacks
    def start_instance(self, result):
        yield self.do_start_instance(result)
        d, self._start_deferred = self._start_deferred, None
        d.callback(result)

    @defer.inlineCallbacks
    def do_start_instance(self, result):
        assert self.state == States.STARTING
        self.state = States.STARTED
        if self.auto_connect_worker and result is True:
            yield self.connect_worker()

    @defer.inlineCallbacks
    def auto_stop(self, result):
        self.auto_stop_flag = result
        if self.auto_stop_flag and self.state == States.STOPPING:
            yield self.stop_instance(True)

    @defer.inlineCallbacks
    def stop_instance(self, result):
        yield self.do_stop_instance()
        d, self._stop_deferred = self._stop_deferred, None
        d.callback(result)

    @defer.inlineCallbacks
    def do_stop_instance(self):
        assert self.state == States.STOPPING
        self.state = States.STOPPED
        self._started_kind = None
        if self.auto_disconnect_worker:
            yield self.disconnect_worker()

    @defer.inlineCallbacks
    def connect_worker(self):
        if self.remote_worker is not None:
            return
        if RemoteWorker is None:
            raise SkipTest("buildbot-worker package is not installed")
        workdir = FilePath(self.case.mktemp())
        workdir.createDirectory()
        self.remote_worker = RemoteWorker(self.worker.name, workdir.path, False)
        yield self.remote_worker.setServiceParent(self.worker)

    @defer.inlineCallbacks
    def disconnect_worker(self):
        yield super().disconnect_worker()
        if self.remote_worker is None:
            return

        self.remote_worker, worker = None, self.remote_worker

        disconnect_master_side_worker(self.worker)

        yield worker.disownServiceParent()

    def setup_kind(self, build):
        if build:
            self._started_kind_deferred = build.render(self.kind)
        else:
            self._started_kind_deferred = self.kind

    @defer.inlineCallbacks
    def get_started_kind(self):
        if self._started_kind_deferred:
            self._started_kind = yield self._started_kind_deferred
            self._started_kind_deferred = None
        return self._started_kind

    def patchBot(self, case, remoteMethod, patch):
        case.patch(BotPbLike, remoteMethod, patch)


class ControllableLatentWorker(AbstractLatentWorker):

    """
    A latent worker that can be controlled by tests.
    """
    builds_may_be_incompatible = True

    def __init__(self, name, controller, **kwargs):
        self._controller = controller
        self._random_password_id = 0
        AbstractLatentWorker.__init__(self, name, None, **kwargs)

    def checkConfig(self, name, _, **kwargs):
        AbstractLatentWorker.checkConfig(
            self, name, None,
            build_wait_timeout=self._controller.build_wait_timeout,
            **kwargs)

    def reconfigService(self, name, _, **kwargs):
        return super().reconfigService(name, self.getRandomPass(),
                                       build_wait_timeout=self._controller.build_wait_timeout,
                                       **kwargs)

    def _generate_random_password(self):
        self._random_password_id += 1
        return f'password_{self._random_password_id}'

    @defer.inlineCallbacks
    def isCompatibleWithBuild(self, build_props):
        if self._controller.state == States.STOPPED:
            return True

        requested_kind = yield build_props.render((self._controller.kind))
        curr_kind = yield self._controller.get_started_kind()
        return requested_kind == curr_kind

    def start_instance(self, build):
        self._controller.setup_kind(build)

        assert self._controller.state == States.STOPPED
        self._controller.state = States.STARTING

        if self._controller.auto_start_flag:
            self._controller.do_start_instance(True)
            return defer.succeed(True)

        self._controller._start_deferred = defer.Deferred()
        return self._controller._start_deferred

    @defer.inlineCallbacks
    def stop_instance(self, fast):
        assert self._controller.state == States.STARTED
        self._controller.state = States.STOPPING

        if self._controller.auto_stop_flag:
            yield self._controller.do_stop_instance()
            return True
        self._controller._stop_deferred = defer.Deferred()
        return (yield self._controller._stop_deferred)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from parameterized import parameterized

import mock

from twisted.internet import defer
from twisted.python import runtime
from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.logging import LoggingMixin
from buildbot.util.runprocess import RunProcess

# windows returns rc 1, because exit status cannot indicate "signalled";
# posix returns rc -1 for "signalled"
FATAL_RC = -1
if runtime.platformType == 'win32':
    FATAL_RC = 1


class TestRunProcess(TestReactorMixin, LoggingMixin, unittest.TestCase):

    FAKE_PID = 1234

    def setUp(self):
        self.setup_test_reactor()
        self.setUpLogging()
        self.process = None
        self.reactor.spawnProcess = self.fake_spawn_process

    def fake_spawn_process(self, pp, command, args, env, workdir, usePTY=False):
        self.assertIsNone(self.process)
        self.pp = pp
        self.pp.transport = mock.Mock()
        self.process = mock.Mock()
        self.process.pid = self.FAKE_PID
        self.process_spawned_args = (command, args, env, workdir)
        return self.process

    def run_process(self, command, override_kill_success=True, override_is_dead=True, **kwargs):
        self.run_process_obj = RunProcess(self.reactor, command, '/workdir', **kwargs)
        self.run_process_obj.get_os_env = lambda: {'OS_ENV': 'value'}
        self.run_process_obj.send_signal = mock.Mock(side_effect=lambda sig: override_kill_success)
        self.run_process_obj.is_dead = mock.Mock(side_effect=lambda: override_is_dead)
        return self.run_process_obj.start()

    def end_process(self, signal=None, rc=0):
        reason = mock.Mock()
        reason.value.signal = signal
        reason.value.exitCode = rc
        self.pp.processEnded(reason)

    @defer.inlineCallbacks
    def test_no_output(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=False)
        self.assertEqual(self.process_spawned_args,
                         ('cmd', ['cmd'], {'OS_ENV': 'value', 'PWD': os.path.abspath('/workdir')},
                          '/workdir'))

        self.pp.connectionMade()
        self.assertFalse(d.called)
        self.end_process()
        self.assertTrue(d.called)

        res = yield d
        self.assertEqual(res, (0, b''))

    @defer.inlineCallbacks
    def test_env_new_kv(self):
        d = self.run_process(['cmd'], collect_stdout=False, collect_stderr=False,
                             env={'custom': 'custom-value'})
        self.assertEqual(self.process_spawned_args,
                         ('cmd', ['cmd'], {'OS_ENV': 'value', 'PWD': os.path.abspath('/workdir'),
                                           'custom': 'custom-value'}, '/workdir'))

        self.pp.connectionMade()
        self.end_process()

        res = yield d
        self.assertEqual(res, 0)

    @defer.inlineCallbacks
    def test_env_overwrite_os_kv(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=False,
                             env={'OS_ENV': 'custom-value'})
        self.assertEqual(self.process_spawned_args,
                         ('cmd', ['cmd'], {'OS_ENV': 'custom-value',
                                           'PWD': os.path.abspath('/workdir')},
                          '/workdir'))

        self.pp.connectionMade()
        self.end_process()

        res = yield d
        self.assertEqual(res, (0, b''))

    @defer.inlineCallbacks
    def test_env_remove_os_kv(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=False,
                             env={'OS_ENV': None})
        self.assertEqual(self.process_spawned_args,
                         ('cmd', ['cmd'], {'PWD': os.path.abspath('/workdir')}, '/workdir'))

        self.pp.connectionMade()
        self.end_process()

        res = yield d
        self.assertEqual(res, (0, b''))

    @defer.inlineCallbacks
    def test_collect_nothing(self):
        d = self.run_process(['cmd'], collect_stdout=False, collect_stderr=False)

        self.pp.connectionMade()
        self.pp.transport.write.assert_not_called()
        self.pp.transport.closeStdin.assert_called()

        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')

        self.assertFalse(d.called)
        self.end_process()
        self.assertTrue(d.called)

        res = yield d
        self.assertEqual(res, 0)

    @defer.inlineCallbacks
    def test_collect_stdout_no_stderr(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=False)

        self.pp.connectionMade()
        self.pp.transport.write.assert_not_called()
        self.pp.transport.closeStdin.assert_called()

        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')

        self.assertFalse(d.called)
        self.end_process()
        self.assertTrue(d.called)

        res = yield d
        self.assertEqual(res, (0, b'stdout_data'))

    @defer.inlineCallbacks
    def test_collect_stdout_with_stdin(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=False,
                             initial_stdin=b'stdin')

        self.pp.connectionMade()
        self.pp.transport.write.assert_called_with(b'stdin')
        self.pp.transport.closeStdin.assert_called()

        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')
        self.end_process()

        res = yield d
        self.assertEqual(res, (0, b'stdout_data'))

    @defer.inlineCallbacks
    def test_collect_stdout_and_stderr(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True)

        self.pp.connectionMade()
        self.pp.transport.write.assert_not_called()
        self.pp.transport.closeStdin.assert_called()

        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')
        self.end_process()

        res = yield d
        self.assertEqual(res, (0, b'stdout_data', b'stderr_data'))

    @defer.inlineCallbacks
    def test_process_failed_with_rc(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True)

        self.pp.connectionMade()
        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')
        self.end_process(rc=1)

        res = yield d
        self.assertEqual(res, (1, b'stdout_data', b'stderr_data'))

    @defer.inlineCallbacks
    def test_process_failed_with_signal(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True)

        self.pp.connectionMade()
        self.pp.outReceived(b'stdout_data')
        self.pp.errReceived(b'stderr_data')
        self.end_process(signal='SIGILL')

        res = yield d
        self.assertEqual(res, (-1, b'stdout_data', b'stderr_data'))

    @parameterized.expand([
        ('too_short_time_no_output', 0, 4.9, False, False, False),
        ('too_short_time_with_output', 0, 4.9, False, True, True),
        ('timed_out_no_output', 0, 5.1, True, False, False),
        ('timed_out_with_output', 0, 5.1, True, True, True),
        ('stdout_prevented_timeout', 1.0, 4.9, False, True, False),
        ('stderr_prevented_timeout', 1.0, 4.9, False, False, True),
        ('timed_out_after_extra_output', 1.0, 5.1, True, True, True),
    ])
    @defer.inlineCallbacks
    def test_io_timeout(self, name, wait1, wait2, timed_out, had_stdout, had_stderr):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True, io_timeout=5)

        self.pp.connectionMade()
        self.reactor.advance(wait1)
        if had_stdout:
            self.pp.outReceived(b'stdout_data')
        if had_stderr:
            self.pp.errReceived(b'stderr_data')
        self.reactor.advance(wait2)

        self.assertFalse(d.called)
        self.end_process()
        self.assertTrue(d.called)

        if timed_out:
            self.run_process_obj.send_signal.assert_called_with('TERM')
        else:
            self.run_process_obj.send_signal.assert_not_called()

        res = yield d
        self.assertEqual(res, (FATAL_RC if timed_out else 0,
                               b'stdout_data' if had_stdout else b'',
                               b'stderr_data' if had_stderr else b''))

    @parameterized.expand([
        ('too_short_time', 4.9, False),
        ('timed_out', 5.1, True),
    ])
    @defer.inlineCallbacks
    def test_runtime_timeout(self, name, wait, timed_out):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True, runtime_timeout=5)

        self.pp.connectionMade()
        self.reactor.advance(wait)

        self.assertFalse(d.called)
        self.end_process()
        self.assertTrue(d.called)

        if timed_out:
            self.run_process_obj.send_signal.assert_called_with('TERM')
        else:
            self.run_process_obj.send_signal.assert_not_called()

        res = yield d
        self.assertEqual(res, (FATAL_RC if timed_out else 0, b'', b''))

    @defer.inlineCallbacks
    def test_runtime_timeout_failing_to_kill(self):
        d = self.run_process(['cmd'], collect_stdout=True, collect_stderr=True, runtime_timeout=5,
                             sigterm_timeout=5, override_is_dead=False)

        self.pp.connectionMade()
        self.reactor.advance(5.1)
        self.run_process_obj.send_signal.assert_called_with('TERM')
        self.reactor.advance(5.1)
        self.run_process_obj.send_signal.assert_called_with('KILL')
        self.reactor.advance(5.1)

        self.assertTrue(d.called)
        self.end_process()

        with self.assertRaises(RuntimeError):
            yield d

        self.assertLogged("attempted to kill process, but it wouldn't die")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import connector
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.util import unicode2bytes
from buildbot.www import graphql

try:
    import graphql as graphql_core
except ImportError:
    graphql_core = None


class V3RootResource(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    if not graphql_core:
        skip = "graphql is required for V3RootResource tests"

    def setUp(self):
        self.patch(connector.DataConnector, 'submodules', [])
        self.setup_test_reactor(use_asyncio=True)
        self.master = self.make_master(url="http://server/path/", wantGraphql=True)
        self.master.config.www["graphql"] = {"debug": True}
        self.rsrc = graphql.V3RootResource(self.master)
        self.rsrc.reconfigResource(self.master.config)

    def assertSimpleError(self, message_or_error, responseCode):
        if isinstance(message_or_error, list):
            errors = message_or_error
        else:
            errors = [{"message": message_or_error}]
        content = json.dumps({"data": None, "errors": errors})
        self.assertRequest(content=unicode2bytes(content), responseCode=responseCode)

    def assertResult(self, result):
        content = json.dumps({"data": result, "errors": None})
        self.assertRequest(content=unicode2bytes(content), responseCode=200)

    @defer.inlineCallbacks
    def test_failure(self):
        self.master.graphql.query = mock.Mock(
            return_value=defer.fail(RuntimeError("oh noes"))
        )
        yield self.render_resource(
            self.rsrc,
            b"/?query={builders{name}}",
        )
        self.assertSimpleError("internal error - see logs", 500)
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def test_invalid_http_method(self):
        yield self.render_resource(self.rsrc, b"/", method=b"PATCH")
        self.assertSimpleError("invalid HTTP method", 400)

    # https://graphql.org/learn/serving-over-http/#get-request
    @defer.inlineCallbacks
    def test_get_query(self):
        yield self.render_resource(
            self.rsrc,
            b"/?query={tests{testid}}",
        )
        self.assertResult(
            {
                "tests": [
                    {"testid": 13},
                    {"testid": 14},
                    {"testid": 15},
                    {"testid": 16},
                    {"testid": 17},
                    {"testid": 18},
                    {"testid": 19},
                    {"testid": 20},
                ]
            }
        )

    @defer.inlineCallbacks
    def test_get_query_item(self):
        yield self.render_resource(
            self.rsrc,
            b"/?query={test(testid:13){testid, info}}",
        )
        self.assertResult({"test": {"testid": 13, "info": "ok"}})

    @defer.inlineCallbacks
    def test_get_query_subresource(self):
        yield self.render_resource(
            self.rsrc,
            b"/?query={test(testid:13){testid, info, steps { info }}}",
        )
        self.assertResult(
            {
                "test": {
                    "testid": 13,
                    "info": "ok",
                    "steps": [{"info": "ok"}, {"info": "failed"}],
                }
            }
        )

    @defer.inlineCallbacks
    def test_get_query_items_result_spec(self):
        yield self.render_resource(
            self.rsrc,
            b"/?query={tests(testid__gt:18){testid, info}}",
        )
        self.assertResult(
            {"tests": [{"testid": 19, "info": "todo"}, {"testid": 20, "info": "error"}]}
        )

    @defer.inlineCallbacks
    def test_get_noquery(self):
        yield self.render_resource(
            self.rsrc,
            b"/",
        )
        self.assertSimpleError("GET request must contain a 'query' parameter", 400)

    # https://graphql.org/learn/serving-over-http/#post-request
    @defer.inlineCallbacks
    def test_post_query_graphql_content(self):
        yield self.render_resource(
            self.rsrc,
            method=b"POST",
            content=b"{tests{testid}}",
            content_type=b"application/graphql",
        )
        self.assertResult(
            {
                "tests": [
                    {"testid": 13},
                    {"testid": 14},
                    {"testid": 15},
                    {"testid": 16},
                    {"testid": 17},
                    {"testid": 18},
                    {"testid": 19},
                    {"testid": 20},
                ]
            }
        )

    @defer.inlineCallbacks
    def test_post_query_json_content(self):
        query = {"query": "{tests{testid}}"}
        yield self.render_resource(
            self.rsrc,
            method=b"POST",
            content=json.dumps(query).encode(),
            content_type=b"application/json",
        )
        self.assertResult(
            {
                "tests": [
                    {"testid": 13},
                    {"testid": 14},
                    {"testid": 15},
                    {"testid": 16},
                    {"testid": 17},
                    {"testid": 18},
                    {"testid": 19},
                    {"testid": 20},
                ]
            }
        )

    @defer.inlineCallbacks
    def test_post_query_json_content_operationName(self):
        query = {
            "query": "query foo {tests{testid}} query bar {tests{name}}",
            "operationName": "fsoo",
        }
        yield self.render_resource(
            self.rsrc,
            method=b"POST",
            content=json.dumps(query).encode(),
            content_type=b"application/json",
        )
        self.assertSimpleError("json request unsupported fields: operationName", 400)

    @defer.inlineCallbacks
    def test_post_query_json_badcontent_type(self):

        yield self.render_resource(
            self.rsrc, method=b"POST", content=b"foo", content_type=b"application/foo"
        )
        self.assertSimpleError("unsupported content-type: application/foo", 400)

    @defer.inlineCallbacks
    def test_post_query_json_nocontent_type(self):

        yield self.render_resource(self.rsrc, method=b"POST")
        self.assertSimpleError("no content-type", 400)

    @defer.inlineCallbacks
    def test_get_bad_query(self):
        yield self.render_resource(
            self.rsrc,
            b"/?query={notexistant{id}}",
        )
        self.assertSimpleError(
            [
                {
                    "message": "Cannot query field 'notexistant' on type 'Query'.",
                    "locations": [{"line": 1, "column": 2}],
                }
            ],
            200,
        )


class DisabledV3RootResource(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    if not graphql_core:
        skip = "graphql is required for V3RootResource tests"

    def setUp(self):
        self.setup_test_reactor()
        self.master = self.make_master(url="http://server/path/")
        self.rsrc = graphql.V3RootResource(self.master)
        self.rsrc.reconfigResource(self.master.config)

    @defer.inlineCallbacks
    def test_basic_disabled(self):
        yield self.render_resource(self.rsrc, b"/")
        self.assertRequest(
            content=unicode2bytes(
                json.dumps(
                    {"data": None, "errors": [{"message": "graphql not enabled"}]}
                )
            ),
            responseCode=501,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#

""" This module holds configurators, which helps setup schedulers, builders, steps,
    for a very specific purpose.
    Higher level interfaces to buildbot configurations components.
"""

from zope.interface import implementer

from buildbot.interfaces import IConfigurator


@implementer(IConfigurator)
class ConfiguratorBase:
    """
        I provide base helper methods for configurators
    """
    def __init__(self):
        pass

    def configure(self, config_dict):
        self.config_dict = c = config_dict
        if 'schedulers' not in c:
            c['schedulers'] = []
        self.schedulers = c['schedulers']
        if 'protocols' not in c:
            c['protocols'] = {}
        self.protocols = c['protocols']
        if 'builders' not in c:
            c['builders'] = []
        self.builders = c['builders']
        if 'workers' not in c:
            c['workers'] = []
        self.workers = c['workers']



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot import interfaces
from buildbot.process.properties import Interpolate
from buildbot.process.properties import Properties
from buildbot.process.properties import Property
from buildbot.test.fake import docker
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.worker import docker as dockerworker


class TestDockerLatentWorker(unittest.TestCase, TestReactorMixin):

    @defer.inlineCallbacks
    def setupWorker(self, *args, **kwargs):
        self.patch(dockerworker, 'docker', docker)
        worker = dockerworker.DockerLatentWorker(*args, **kwargs)
        master = fakemaster.make_master(self, wantData=True)
        fakemaster.master = master
        worker.setServiceParent(master)
        yield master.startService()
        self.addCleanup(master.stopService)
        return worker

    def setUp(self):
        self.setup_test_reactor()

        self.build = Properties(
            image='busybox:latest', builder='docker_worker', distro='wheezy')
        self.build2 = Properties(
            image='busybox:latest', builder='docker_worker2', distro='wheezy')
        self.patch(dockerworker, 'client', docker)
        docker.Client.containerCreated = False
        docker.Client.start_exception = None

    @defer.inlineCallbacks
    def test_constructor_nodocker(self):
        self.patch(dockerworker, 'client', None)
        with self.assertRaises(config.ConfigErrors):
            yield self.setupWorker('bot', 'pass', 'unix://tmp.sock', 'debian:wheezy', [])

    @defer.inlineCallbacks
    def test_constructor_noimage_nodockerfile(self):
        with self.assertRaises(config.ConfigErrors):
            yield self.setupWorker('bot', 'pass', 'http://localhost:2375')

    @defer.inlineCallbacks
    def test_constructor_noimage_dockerfile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'http://localhost:2375', dockerfile="FROM ubuntu")
        self.assertEqual(bs.dockerfile, "FROM ubuntu")
        self.assertEqual(bs.image, None)

    @defer.inlineCallbacks
    def test_constructor_image_nodockerfile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'http://localhost:2375', image="myworker")
        self.assertEqual(bs.dockerfile, None)
        self.assertEqual(bs.image, 'myworker')

    @defer.inlineCallbacks
    def test_constructor_minimal(self):
        # Minimal set of parameters
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker')
        self.assertEqual(bs.workername, 'bot')
        self.assertEqual(bs.password, 'pass')
        self.assertEqual(bs.docker_host, 'tcp://1234:2375')
        self.assertEqual(bs.client_args, {})
        self.assertEqual(bs.image, 'worker')
        self.assertEqual(bs.command, [])

    @defer.inlineCallbacks
    def test_builds_may_be_incompatible(self):
        # Minimal set of parameters
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker')
        self.assertEqual(bs.builds_may_be_incompatible, True)

    @defer.inlineCallbacks
    def test_contruction_minimal_docker_py(self):
        self.patch(dockerworker, 'docker_py_version', 1.10)
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker')
        yield bs.start_instance(self.build)
        client = docker.APIClient.latest
        self.assertEqual(client.called_class_name, "Client")
        client = docker.Client.latest
        self.assertNotEqual(client.called_class_name, "APIClient")

    @defer.inlineCallbacks
    def test_contruction_minimal_docker(self):
        self.patch(dockerworker, 'docker_py_version', 2.0)
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker')
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(client.called_class_name, "APIClient")
        client = docker.APIClient.latest
        self.assertNotEqual(client.called_class_name, "Client")

    @defer.inlineCallbacks
    def test_constructor_nopassword(self):
        # when no password, it is created automatically
        bs = yield self.setupWorker('bot', None, 'tcp://1234:2375', 'worker')
        self.assertEqual(bs.workername, 'bot')
        self.assertEqual(len(bs.password), 20)

    @defer.inlineCallbacks
    def test_constructor_all_docker_parameters(self):
        # Volumes have their own tests
        bs = yield self.setupWorker('bot', 'pass', 'unix:///var/run/docker.sock', 'worker_img',
                                    ['/bin/sh'],
                                    dockerfile="FROM ubuntu", version='1.9', tls=True,
                                    hostconfig={'network_mode': 'fake',
                                                'dns': ['1.1.1.1', '1.2.3.4']},
                                    custom_context=False, buildargs=None,
                                    encoding='gzip')
        self.assertEqual(bs.workername, 'bot')
        self.assertEqual(bs.docker_host, 'unix:///var/run/docker.sock')
        self.assertEqual(bs.password, 'pass')
        self.assertEqual(bs.image, 'worker_img')
        self.assertEqual(bs.command, ['/bin/sh'])
        self.assertEqual(bs.dockerfile, "FROM ubuntu")
        self.assertEqual(bs.volumes, [])
        self.assertEqual(bs.client_args, {'version': '1.9', 'tls': True})
        self.assertEqual(
            bs.hostconfig, {'network_mode': 'fake', 'dns': ['1.1.1.1', '1.2.3.4']})
        self.assertFalse(bs.custom_context)
        self.assertEqual(bs.buildargs, None)
        self.assertEqual(bs.encoding, 'gzip')

    @defer.inlineCallbacks
    def test_constructor_host_config_build(self):
        # Volumes have their own tests
        bs = yield self.setupWorker('bot', 'pass', 'unix:///var/run/docker.sock', 'worker_img',
                                    ['/bin/sh'],
                                    dockerfile="FROM ubuntu",
                                    volumes=["/tmp:/tmp:ro"],
                                    hostconfig={'network_mode': 'fake',
                                                'dns': ['1.1.1.1', '1.2.3.4']},
                                    custom_context=False, buildargs=None,
                                    encoding='gzip')
        yield bs.start_instance(self.build)
        client = docker.APIClient.latest
        expected = {
            'network_mode': 'fake',
            'dns': ['1.1.1.1', '1.2.3.4'],
            'binds': ['/tmp:/tmp:ro'],
        }
        if dockerworker.docker_py_version >= 2.2:
            expected['init'] = True
        self.assertEqual(client.call_args_create_host_config, [expected])

    @defer.inlineCallbacks
    def test_constructor_host_config_build_set_init(self):
        # Volumes have their own tests
        bs = yield self.setupWorker('bot', 'pass', 'unix:///var/run/docker.sock', 'worker_img',
                                    ['/bin/sh'],
                                    dockerfile="FROM ubuntu",
                                    volumes=["/tmp:/tmp:ro"],
                                    hostconfig={'network_mode': 'fake',
                                                'dns': ['1.1.1.1', '1.2.3.4'],
                                                'init': False},
                                    custom_context=False, buildargs=None,
                                    encoding='gzip')
        yield bs.start_instance(self.build)
        client = docker.APIClient.latest
        self.assertEqual(client.call_args_create_host_config, [
            {'network_mode': 'fake',
             'dns': ['1.1.1.1', '1.2.3.4'],
             'init': False,
             'binds': ['/tmp:/tmp:ro'],
             }
        ])

    @defer.inlineCallbacks
    def test_start_instance_docker_host_renderable(self):
        bs = yield self.setupWorker('bot', 'pass',
                                    docker_host=Interpolate('tcp://value-%(prop:builder)s'),
                                    image='worker')
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(client.base_url, 'tcp://value-docker_worker')

    @defer.inlineCallbacks
    def test_start_instance_volume_renderable(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'worker', ['bin/bash'],
            volumes=[Interpolate('/data:/worker/%(kw:builder)s/build',
                                 builder=Property('builder'))])
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(len(client.call_args_create_container), 1)
        self.assertEqual(client.call_args_create_container[0]['volumes'],
                         ['/worker/docker_worker/build'])

    @defer.inlineCallbacks
    def test_start_instance_hostconfig_renderable(self):
        bs = yield self.setupWorker('bot', 'pass', docker_host='tcp://1234:2375', image='worker',
                                    hostconfig={'prop': Interpolate('value-%(kw:builder)s',
                                                                    builder=Property('builder'))})
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(len(client.call_args_create_container), 1)

        expected = {'prop': 'value-docker_worker', 'binds': []}
        if dockerworker.docker_py_version >= 2.2:
            expected['init'] = True
        self.assertEqual(client.call_args_create_host_config, [expected])

    @defer.inlineCallbacks
    def test_interpolate_renderables_for_new_build(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'worker', ['bin/bash'],
            volumes=[Interpolate('/data:/worker/%(kw:builder)s/build',
                                 builder=Property('builder'))])
        yield bs.start_instance(self.build)
        docker.Client.containerCreated = True
        # the worker recreates the (mock) client on every action, clearing the containers
        # but stop_instance only works if the there is a docker container running
        yield bs.stop_instance()
        self.assertTrue((yield bs.isCompatibleWithBuild(self.build2)))

    @defer.inlineCallbacks
    def test_reject_incompatible_build_while_running(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'worker', ['bin/bash'],
            volumes=[Interpolate('/data:/worker/%(kw:builder)s/build',
                                 builder=Property('builder'))])
        yield bs.start_instance(self.build)
        self.assertFalse((yield bs.isCompatibleWithBuild(self.build2)))

    @defer.inlineCallbacks
    def test_volume_no_suffix(self):
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker', ['bin/bash'],
                                    volumes=['/src/webapp:/opt/webapp'])
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(len(client.call_args_create_container), 1)
        self.assertEqual(len(client.call_args_create_host_config), 1)
        self.assertEqual(client.call_args_create_container[0]['volumes'],
                         ['/opt/webapp'])
        self.assertEqual(client.call_args_create_host_config[0]['binds'],
                         ["/src/webapp:/opt/webapp"])

    @defer.inlineCallbacks
    def test_volume_ro_rw(self):
        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'worker', ['bin/bash'],
                                    volumes=['/src/webapp:/opt/webapp:ro',
                                             '~:/backup:rw'])
        yield bs.start_instance(self.build)
        client = docker.Client.latest
        self.assertEqual(len(client.call_args_create_container), 1)
        self.assertEqual(len(client.call_args_create_host_config), 1)
        self.assertEqual(client.call_args_create_container[0]['volumes'],
                         ['/opt/webapp', '/backup'])
        self.assertEqual(client.call_args_create_host_config[0]['binds'],
                         ['/src/webapp:/opt/webapp:ro', '~:/backup:rw'])

    @defer.inlineCallbacks
    def test_volume_bad_format(self):
        with self.assertRaises(config.ConfigErrors):
            yield self.setupWorker('bot', 'pass', 'http://localhost:2375',
                                   image="worker",
                                   volumes=['abcd=efgh'])

    @defer.inlineCallbacks
    def test_volume_bad_format_renderable(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'http://localhost:2375', image="worker",
            volumes=[Interpolate('/data==/worker/%(kw:builder)s/build',
                                 builder=Property('builder'))])
        with self.assertRaises(config.ConfigErrors):
            yield bs.start_instance(self.build)

    @defer.inlineCallbacks
    def test_start_instance_image_no_version(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'busybox', ['bin/bash'])
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'busybox')

    @defer.inlineCallbacks
    def test_start_instance_image_right_version(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'busybox:latest', ['bin/bash'])
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'busybox:latest')

    @defer.inlineCallbacks
    def test_start_instance_image_wrong_version(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'busybox:123', ['bin/bash'])
        with self.assertRaises(interfaces.LatentWorkerCannotSubstantiate):
            yield bs.start_instance(self.build)

    @defer.inlineCallbacks
    def test_start_instance_image_renderable(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', Property('image'), ['bin/bash'])
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'busybox:latest')

    @defer.inlineCallbacks
    def test_start_instance_noimage_nodockerfile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'customworker', ['bin/bash'])
        with self.assertRaises(interfaces.LatentWorkerCannotSubstantiate):
            yield bs.start_instance(self.build)

    @defer.inlineCallbacks
    def test_start_instance_image_and_dockefile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'customworker', dockerfile='BUG')
        with self.assertRaises(interfaces.LatentWorkerCannotSubstantiate):
            yield bs.start_instance(self.build)

    @defer.inlineCallbacks
    def test_start_instance_noimage_gooddockerfile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'customworker', dockerfile='FROM debian:wheezy')
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'customworker')

    @defer.inlineCallbacks
    def test_start_instance_noimage_pull(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'alpine:latest', autopull=True)
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'alpine:latest')

    @defer.inlineCallbacks
    def test_start_instance_image_pull(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest', autopull=True)
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')
        client = docker.Client.latest
        self.assertEqual(client._pullCount, 0)

    @defer.inlineCallbacks
    def test_start_instance_image_alwayspull(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest', autopull=True, alwaysPull=True)
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')
        client = docker.Client.latest
        self.assertEqual(client._pullCount, 1)

    @defer.inlineCallbacks
    def test_start_instance_image_noauto_alwayspull(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest', autopull=False, alwaysPull=True)
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')
        client = docker.Client.latest
        self.assertEqual(client._pullCount, 0)

    @defer.inlineCallbacks
    def test_start_instance_noimage_renderabledockerfile(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'customworker',
            dockerfile=Interpolate('FROM debian:%(kw:distro)s',
                                   distro=Property('distro')))
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'customworker')

    @defer.inlineCallbacks
    def test_start_instance_custom_context_and_buildargs(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest',
            dockerfile=Interpolate('FROM debian:latest'), custom_context=True,
            buildargs={'sample_arg1': 'test_val1'})
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')

    @defer.inlineCallbacks
    def test_start_instance_custom_context_no_buildargs(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest',
            dockerfile=Interpolate('FROM debian:latest'),
            custom_context=True)
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')

    @defer.inlineCallbacks
    def test_start_instance_buildargs_no_custom_context(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'tcp://1234:2375', 'tester:latest',
            dockerfile=Interpolate('FROM debian:latest'),
            buildargs={'sample_arg1': 'test_val1'})
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'tester:latest')

    @defer.inlineCallbacks
    def test_start_worker_but_already_created_with_same_name(self):
        bs = yield self.setupWorker(
            'existing', 'pass', 'tcp://1234:2375', 'busybox:latest', ['bin/bash'])
        _, name = yield bs.start_instance(self.build)
        self.assertEqual(name, 'busybox:latest')

    @defer.inlineCallbacks
    def test_start_instance_docker_client_start_exception(self):
        msg = 'The container operating system does not match the host operating system'
        docker.Client.start_exception = docker.errors.APIError(msg)

        bs = yield self.setupWorker('bot', 'pass', 'tcp://1234:2375', 'busybox:latest',
                                    ['bin/bash'])
        with self.assertRaises(interfaces.LatentWorkerCannotSubstantiate):
            yield bs.start_instance(self.build)

    @defer.inlineCallbacks
    def test_constructor_hostname(self):
        bs = yield self.setupWorker(
            'bot', 'pass', 'http://localhost:2375',
            image="myworker_image", hostname="myworker_hostname")
        self.assertEqual(bs.hostname, 'myworker_hostname')


class testDockerPyStreamLogs(unittest.TestCase):

    def compare(self, result, log):
        self.assertEqual(result,
                         list(dockerworker._handle_stream_line(log)))

    def testEmpty(self):
        self.compare([], '{"stream":"\\n"}\r\n')

    def testOneLine(self):
        self.compare(
            [" ---> Using cache"], '{"stream":" ---\\u003e Using cache\\n"}\r\n')

    def testMultipleLines(self):
        self.compare(["Fetched 8298 kB in 3s (2096 kB/s)", "Reading package lists..."],
                     '{"stream": "Fetched 8298 kB in 3s (2096 kB/s)\\nReading '
                     'package lists..."}\r\n')

    def testError(self):
        self.compare(["ERROR: The command [/bin/sh -c apt-get update && apt-get install -y"
                      "    python-dev    python-pip] returned a non-zero code: 127"],
                     '{"errorDetail": {"message": "The command [/bin/sh -c apt-get update && '
                     'apt-get install -y    python-dev    python-pip] returned a '
                     'non-zero code: 127"},'
                     ' "error": "The command [/bin/sh -c apt-get update && apt-get install -y'
                     '    python-dev    python-pip] returned a non-zero code: 127"}\r\n')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import warnings

from twisted.trial import unittest

from buildbot.warnings import DeprecatedApiWarning


def deprecatedImport(fn):
    def wrapper(self):
        fn(self)
        warnings = self.flushWarnings()
        # on older Pythons, this warning appears twice, so use collapse it
        if len(warnings) == 2 and warnings[0] == warnings[1]:
            del warnings[1]
        self.assertEqual(len(warnings), 1, f"got: {repr(warnings)}")
        self.assertEqual(warnings[0]['category'], DeprecatedApiWarning)
    return wrapper


class OldImportPaths(unittest.TestCase):

    """
    Test that old, deprecated import paths still work.
    """

    def test_scheduler_Scheduler(self):
        from buildbot.scheduler import Scheduler
        assert Scheduler

    def test_schedulers_basic_Scheduler(self):
        # renamed to basic.SingleBranchScheduler
        from buildbot.schedulers.basic import Scheduler
        assert Scheduler

    def test_scheduler_AnyBranchScheduler(self):
        from buildbot.scheduler import AnyBranchScheduler
        assert AnyBranchScheduler

    def test_scheduler_basic_Dependent(self):
        from buildbot.schedulers.basic import Dependent
        assert Dependent

    def test_scheduler_Dependent(self):
        from buildbot.scheduler import Dependent
        assert Dependent

    def test_scheduler_Periodic(self):
        from buildbot.scheduler import Periodic
        assert Periodic

    def test_scheduler_Nightly(self):
        from buildbot.scheduler import Nightly
        assert Nightly

    def test_scheduler_Triggerable(self):
        from buildbot.scheduler import Triggerable
        assert Triggerable

    def test_scheduler_Try_Jobdir(self):
        from buildbot.scheduler import Try_Jobdir
        assert Try_Jobdir

    def test_scheduler_Try_Userpass(self):
        from buildbot.scheduler import Try_Userpass
        assert Try_Userpass

    def test_schedulers_filter_ChangeFilter(self):
        # this was the location of ChangeFilter until 0.8.4
        from buildbot.schedulers.filter import ChangeFilter
        assert ChangeFilter

    def test_process_base_Build(self):
        from buildbot.process.base import Build
        assert Build

    def test_buildrequest_BuildRequest(self):
        from buildbot.buildrequest import BuildRequest
        assert BuildRequest

    def test_process_subunitlogobserver_SubunitShellCommand(self):
        from buildbot.process.subunitlogobserver import SubunitShellCommand
        assert SubunitShellCommand

    def test_steps_source_Source(self):
        from buildbot.steps.source import Source
        assert Source

    def test_buildstep_remotecommand(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecatedApiWarning)
            warnings.simplefilter("ignore", DeprecationWarning)
            from buildbot.process.buildstep import RemoteCommand, \
                LoggedRemoteCommand, RemoteShellCommand
            assert RemoteCommand
            assert LoggedRemoteCommand
            assert RemoteShellCommand

    def test_buildstep_logobserver(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecatedApiWarning)
            warnings.simplefilter("ignore", DeprecationWarning)
            from buildbot.process.buildstep import LogObserver, \
                LogLineObserver, OutputProgressObserver
        assert LogObserver
        assert LogLineObserver
        assert OutputProgressObserver



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Alternatively, you can use and copy this module under the MIT License
# Copyright Buildbot Team Members

import asyncio
import inspect
from asyncio import base_events
from asyncio import events

from twisted.internet import defer


def deferred_await(self):
    # if a deferred is awaited from a asyncio loop context, we must return
    # the future wrapper, but if it is awaited from normal twisted loop
    # we must return self.
    if isinstance(asyncio.get_event_loop(), AsyncIOLoopWithTwisted):
        return self.asFuture(asyncio.get_event_loop())
    return self


defer.Deferred.__await__ = deferred_await


def as_deferred(f):
    return asyncio.get_event_loop().as_deferred(f)


def as_future(d):
    return d.asFuture(asyncio.get_event_loop())


class AsyncIOLoopWithTwisted(base_events.BaseEventLoop):
    """
    Minimal asyncio loop for Buildbot asyncio only dependencies
    as of now, only graphql is needing asyncio loop

    As of now, it can only run basic coroutines, no network operation is supported
    But this could be implemented as needed
    """

    def __init__(self, reactor):
        self._running = False
        self._reactor = reactor
        super().__init__()
        self._running = True

    def start(self):
        self._running = True
        events._set_running_loop(self)

    def stop(self):
        self._running = False
        events._set_running_loop(None)

    def is_running(self):
        return self._running

    def call_soon(self, callback, *args, context=None):
        handle = events.Handle(callback, args, self, context)

        self._reactor.callLater(0, handle._run)
        return handle

    def call_soon_threadsafe(self, callback, *args, context=None):
        handle = events.Handle(callback, args, self, context)

        self._reactor.callFromThread(handle._run)
        return handle

    def time(self):
        # we delegate timekeeping to the reactor so that it can be faked
        return self._reactor.seconds()

    def call_at(self, when, callback, *args, context=None):
        handle = events.Handle(callback, args, self, context)

        # Twisted timers are relatives, contrary to asyncio.
        delay = when - self.time()
        delay = max(delay, 0)
        self._reactor.callLater(delay, handle._run)
        return handle

    def as_deferred(self, thing):
        if isinstance(thing, defer.Deferred):
            return thing

        # check for coroutine objects
        if inspect.isawaitable(thing):
            return defer.Deferred.fromFuture(asyncio.ensure_future(thing))

        return defer.succeed(thing)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.test_result_submitter import TestResultSubmitter


class TestTestResultSubmitter(TestReactorMixin, unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantData=True, wantDb=True)
        yield self.master.startService()

        self.master.db.insertTestData([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88,
                         builderid=88, workerid=47),
            fakedb.Step(id=131, number=132, name='step132', buildid=30),
        ])

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.master.stopService()

    @defer.inlineCallbacks
    def test_complete_empty(self):
        sub = TestResultSubmitter()
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'cat', 'unit')

        setid = sub.get_test_result_set_id()
        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'cat',
            'value_unit': 'unit',
            'tests_passed': None,
            'tests_failed': None,
            'complete': False
        }])

        yield sub.finish()

        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'cat',
            'value_unit': 'unit',
            'tests_passed': None,
            'tests_failed': None,
            'complete': True
        }])

    @defer.inlineCallbacks
    def test_submit_result(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'cat', 'unit')
        sub.add_test_result('1', 'name1')
        yield sub.finish()

        setid = sub.get_test_result_set_id()

        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'cat',
            'value_unit': 'unit',
            'tests_passed': None,
            'tests_failed': None,
            'complete': True
        }])

        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        self.assertEqual(list(results), [{
            'test_resultid': 1002,
            'builderid': 88,
            'test_result_setid': setid,
            'test_name': 'name1',
            'test_code_path': None,
            'duration_ns': None,
            'line': None,
            'value': '1'
        }])

    def filter_results_value_name(self, results):
        return [{'test_name': r['test_name'], 'value': r['value']} for r in results]

    @defer.inlineCallbacks
    def test_submit_result_wrong_argument_types(self):
        sub = TestResultSubmitter()
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'cat', 'unit')

        with self.assertRaises(TypeError):
            sub.add_test_result(1, 'name1')
        with self.assertRaises(TypeError):
            sub.add_test_result('1', test_name=123)
        with self.assertRaises(TypeError):
            sub.add_test_result('1', 'name1', test_code_path=123)
        with self.assertRaises(TypeError):
            sub.add_test_result('1', 'name1', line='123')
        with self.assertRaises(TypeError):
            sub.add_test_result('1', 'name1', duration_ns='123')

    @defer.inlineCallbacks
    def test_batchs_last_batch_full(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'cat', 'unit')
        sub.add_test_result('1', 'name1')
        sub.add_test_result('2', 'name2')
        sub.add_test_result('3', 'name3')
        sub.add_test_result('4', 'name4')
        sub.add_test_result('5', 'name5')
        sub.add_test_result('6', 'name6')
        yield sub.finish()

        setid = sub.get_test_result_set_id()

        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        results = self.filter_results_value_name(results)
        self.assertEqual(results, [
            {'test_name': 'name1', 'value': '1'},
            {'test_name': 'name2', 'value': '2'},
            {'test_name': 'name3', 'value': '3'},
            {'test_name': 'name4', 'value': '4'},
            {'test_name': 'name5', 'value': '5'},
            {'test_name': 'name6', 'value': '6'},
        ])

    @defer.inlineCallbacks
    def test_batchs_last_batch_not_full(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'cat', 'unit')
        sub.add_test_result('1', 'name1')
        sub.add_test_result('2', 'name2')
        sub.add_test_result('3', 'name3')
        sub.add_test_result('4', 'name4')
        sub.add_test_result('5', 'name5')
        yield sub.finish()

        setid = sub.get_test_result_set_id()

        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        results = self.filter_results_value_name(results)
        self.assertEqual(results, [
            {'test_name': 'name1', 'value': '1'},
            {'test_name': 'name2', 'value': '2'},
            {'test_name': 'name3', 'value': '3'},
            {'test_name': 'name4', 'value': '4'},
            {'test_name': 'name5', 'value': '5'},
        ])

    @defer.inlineCallbacks
    def test_counts_pass_fail(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'pass_fail', 'boolean')
        sub.add_test_result('0', 'name1')
        sub.add_test_result('0', 'name2')
        sub.add_test_result('1', 'name3')
        sub.add_test_result('1', 'name4')
        sub.add_test_result('0', 'name5')
        yield sub.finish()

        setid = sub.get_test_result_set_id()
        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'pass_fail',
            'value_unit': 'boolean',
            'tests_passed': 2,
            'tests_failed': 3,
            'complete': True
        }])

    @defer.inlineCallbacks
    def test_counts_pass_fail_invalid_values(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'pass_fail', 'boolean')
        sub.add_test_result('0', 'name1')
        sub.add_test_result('0', 'name2')
        sub.add_test_result('1', 'name3')
        sub.add_test_result('1', 'name4')
        sub.add_test_result('invalid', 'name5')
        yield sub.finish()

        setid = sub.get_test_result_set_id()
        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'pass_fail',
            'value_unit': 'boolean',
            'tests_passed': 2,
            'tests_failed': 2,
            'complete': True
        }])

        # also check whether we preserve the "invalid" values in the database.
        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        results = self.filter_results_value_name(results)
        self.assertEqual(results, [
            {'test_name': 'name1', 'value': '0'},
            {'test_name': 'name2', 'value': '0'},
            {'test_name': 'name3', 'value': '1'},
            {'test_name': 'name4', 'value': '1'},
            {'test_name': 'name5', 'value': 'invalid'},
        ])

        self.flushLoggedErrors(ValueError)

    @defer.inlineCallbacks
    def test_counts_pass_only(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'pass_only', 'some_unit')
        sub.add_test_result('string1', 'name1')
        sub.add_test_result('string2', 'name2')
        sub.add_test_result('string3', 'name3')
        sub.add_test_result('string4', 'name4')
        sub.add_test_result('string5', 'name5')
        yield sub.finish()

        setid = sub.get_test_result_set_id()
        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'pass_only',
            'value_unit': 'some_unit',
            'tests_passed': 5,
            'tests_failed': 0,
            'complete': True
        }])

        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        results = self.filter_results_value_name(results)
        self.assertEqual(results, [
            {'test_name': 'name1', 'value': 'string1'},
            {'test_name': 'name2', 'value': 'string2'},
            {'test_name': 'name3', 'value': 'string3'},
            {'test_name': 'name4', 'value': 'string4'},
            {'test_name': 'name5', 'value': 'string5'},
        ])

        self.flushLoggedErrors(ValueError)

    @defer.inlineCallbacks
    def test_counts_fail_only(self):
        sub = TestResultSubmitter(batch_n=3)
        yield sub.setup_by_ids(self.master, 88, 30, 131, 'desc', 'fail_only', 'some_unit')
        sub.add_test_result('string1', 'name1')
        sub.add_test_result('string2', 'name2')
        sub.add_test_result('string3', 'name3')
        sub.add_test_result('string4', 'name4')
        sub.add_test_result('string5', 'name5')
        yield sub.finish()

        setid = sub.get_test_result_set_id()
        sets = yield self.master.data.get(('builds', 30, 'test_result_sets'))
        self.assertEqual(list(sets), [{
            'test_result_setid': setid,
            'builderid': 88,
            'buildid': 30,
            'stepid': 131,
            'description': 'desc',
            'category': 'fail_only',
            'value_unit': 'some_unit',
            'tests_passed': 0,
            'tests_failed': 5,
            'complete': True
        }])

        results = yield self.master.data.get(('test_result_sets', setid, 'results'))
        results = self.filter_results_value_name(results)
        self.assertEqual(results, [
            {'test_name': 'name1', 'value': 'string1'},
            {'test_name': 'name2', 'value': 'string2'},
            {'test_name': 'name3', 'value': 'string3'},
            {'test_name': 'name4', 'value': 'string4'},
            {'test_name': 'name5', 'value': 'string5'},
        ])

        self.flushLoggedErrors(ValueError)
